// $Id:$   -*-c++-*-
// by Chris Crawford <chris2@lns.mit.edu> 2006/10/02
// plots the figure of merit (p^2*n) along a line in parameter space
// 
// $ root -l p2n_line.C+(25,1e6)
// [] opt_combo();
// [] opt_pair();
// [] kpn=0, dsf=.95, psf=-1; opt_combo(); opt_pair();
// [] opt_single();

#include "TNtuple.h"
#include "TH2.h"
#include "TF1.h"
#include "TFile.h"
#include "TCanvas.h"
#include "TString.h"
#include "TAxis.h"
#include "TGaxis.h"
#include "TROOT.h"
#include "TStyle.h"
#include "TApplication.h"

#include <iostream>
using namespace std;

#define BRANCH(var) float var; TBranch* b_##var=ntp->GetBranch(#var); \
                      ntp->SetBranchAddress(#var,&var)

TH2F *hn, *hpn, *hp2n;
float fom, alpha; // returned by p2n_line_fill
float dsf =-1;	  // -1=std RFSF, .95=strict RFSF, 5% deadtime
float psf =-1;	  // strict polarization for wrap-around neutrons
int   kpn = 2;	  // 2: p2n, 1: pn, -2: p|p|n, 0: p2n or 0 if <0
int icol=kBlue, ipal=1;
// int icol=kRed, ipal=1;

TH1* p2n_line_fill
( TNtuple* ntp, int npt, int nev, float* vpar, float* dpar,
  int nch, char** vch, int ntof=100, bool draw=true, int fit=true,
  float phe3=0.6, float nl=4.0, float rfsf_d=15.868,
  double freq=60./*, double cutoff=15.*/)
{
  // npt = number of points to plot in parameter space between 'par-dpar' to 'par+dpar'
  // par = optimization variable midpoints = {phi_rfsf, phi_i,open_i, ...}
  // dpar= range for optimization, between 'par-dpar' to 'par+dpar'
  // nch = number of choppers (npar=1+2*nch)
  // vch = variable for each chopper, eg. "f1" or "g2"
  // plot= plot p2n histogram+graph
  // fit = 0: don't show fit, 1: fit between -1,1

  // set up ntuple
  BRANCH(l); BRANCH(t_sf); BRANCH(vz_sf); BRANCH(t_h2); BRANCH(p_h2);
  float chop[4]; TBranch* b_chop[4];
  for (int j=0;j<nch;++j) {
    b_chop[j]=ntp->GetBranch(vch[j]);
    ntp->SetBranchAddress(vch[j],&chop[j]);
  }

  // single varying parameter?
  int npar=1+2*nch, ksp=(fit==1)?-2:-1;
  for (int j=0;j<npar;++j)
    if (dpar[j]!=0) ksp = (ksp==-1)? j: -2;

  // set up histograms
  delete hn; delete hpn;
  float min = (ksp<0)? -1 : vpar[ksp]-dpar[ksp];
  float max = (ksp<0)? +1 : vpar[ksp]+dpar[ksp];
  hn =new TH2F("n", "n",  npt,min,max, ntof,0,1);
  hpn=new TH2F("pn","pn", npt,min,max, ntof,0,1);

  // set up parameters
  bool ch_var[nch];
  float alpha_[npt], rfsf_phi[npt], ch_phi[npt][nch], ch_open[npt][nch];
  TAxis ax(npt,-1,1);
  for (int j=0;j<nch;++j)
    ch_var[j] = dpar[1+2*j]!=0 || dpar[2+2*j]!=0;
  for (int i=0;i<npt;++i) {
    alpha_[i] = ax.GetBinCenter(i+1);
    rfsf_phi[i] = (vpar[0] + alpha_[i]*dpar[0])/360.;
    for (int j=0;j<nch;++j) {
      ch_phi [i][j] = vpar[1+2*j] + alpha_[i]*dpar[1+2*j];
      ch_open[i][j] = vpar[2+2*j] + alpha_[i]*dpar[2+2*j];
    }
  }

  // loop over events and fill "n","pn" histograms

  const float ksig=0.07855; //(Amg.cm.Ang)^-1 = 5327b*2200m/s*mn/h
  const float nls=nl*ksig, pnls=phe3*nls;
  const float k_rfsf = 3.141592654 * rfsf_d * freq;
  const float l_15meV=2.3352; //Ang = h/sqrt(2*15meV*mn)

  float tx,pol, tx_i,pol_i, chw, var;

  for (int iev=0;iev<nev;++iev) {
    if (!iev) cout<<" :"<<flush;
    else if (iev%1000000==0) cout<<":"<<flush;
    else if (iev%100000==0) cout<<"."<<flush;

    // read variables
    if (!b_l->GetEntry(iev)) break;
    b_t_sf->GetEntry(iev);
    b_vz_sf->GetEntry(iev);
    b_t_h2->GetEntry(iev);
    b_p_h2->GetEntry(iev);
    for (int j=0;j<nch;++j)
      b_chop[j]->GetEntry(iev);

    // normalized TOF
    t_sf *= freq;
    t_h2 = fmod(t_h2*freq, 1);

    // He3 polarizer
    tx = exp(-nls*l)*cosh(pnls*l);
    pol= tanh(pnls*l);

    // LH2 depolarization
    if (l<l_15meV) pol=0;

    // non-optimized choppers
    for (int j=0;j<nch;++j) {
      if (ch_var[j]) continue;
      chw = chop[j]-ch_phi[0][j];
      if (chw<0) chw += 360;
      if (chw>ch_open[0][j]) { tx=0; break; }
    }
    if (tx==0) continue;

    // non-optimized RFSF, averaged over spin sequence
    static const float spin_seq_ave[]={1, -0.5, 0, 0.5, -1, 0.5, 0, -0.5};
    if (dpar[0]==0) {
      if (dsf<0) {
	pol*= spin_seq_ave[int(fabs(t_sf-rfsf_phi[0]-0.5)+0.5)%8];
	pol*=(1-cos(k_rfsf/vz_sf/(fmod(t_sf-rfsf_phi[0]+10,1)+rfsf_phi[0])))/2.;
      } else {
	if (t_sf<rfsf_phi[0] || t_sf>rfsf_phi[0]+dsf) pol*=psf;
      }
    }

    for (int i=0;i<npt;++i) {

      // optimized choppers
      tx_i=tx;
      for (int j=0;j<nch;++j) {
	if (!ch_var[j]) continue;
	chw = chop[j]-ch_phi[i][j];
	if (chw<0) chw += 360;
	if (chw>ch_open[i][j]) { tx_i=0; break; }
      }
      if (tx_i==0) continue;

      // optimized RFSF
      pol_i=pol;
      if (dpar[0]!=0) {
	if (dsf<0) {
	  pol_i*=spin_seq_ave[int(fabs(t_sf-rfsf_phi[i]-0.5)+0.5)%8];
	  pol_i*=(1-cos(k_rfsf/vz_sf/(fmod(t_sf-rfsf_phi[i]+10,1)+rfsf_phi[i])))/2.;
	} else {
	  if (t_sf<rfsf_phi[i] || t_sf>rfsf_phi[i]+dsf) pol_i*=psf;
	}
      }

      // fill histograms
      var = (ksp<0) ? alpha_[i]: vpar[ksp] + alpha_[i]*dpar[ksp];
      hn ->Fill(var, t_h2, p_h2 * tx_i         / nev*ntof);
      hpn->Fill(var, t_h2, p_h2 * tx_i * pol_i / nev*ntof);

    }
  }

  // calculate FOM=p2n, projection; draw both
  delete hp2n;
  hp2n=(TH2F*)hpn->Clone();
  hp2n->SetName("p2n");
  if (kpn==2) {
    hp2n->Multiply(hpn);
    hp2n->Divide(hn);
  } else {
    for (int i=1;i<=hp2n->GetNbinsX();i++) {
      for (int j=1;j<=hp2n->GetNbinsY();j++) {
	float xpn = hpn->GetBinContent(i,j);
	float xn  = hn->GetBinContent(i,j);
	if (kpn==22) hp2n->SetBinContent(i,j, xn==0? 0: xpn*xpn/xn);
	if (kpn==-2) hp2n->SetBinContent(i,j, xn==0? 0: xpn*fabs(xpn)/xn);
	if (kpn==0)  hp2n->SetBinContent(i,j, xpn>0&&xn>0? xpn*xpn/xn : 0);
      }
    }
  }

  TH1* hfom = hp2n->ProjectionX();
  hfom->Scale(1./ntof);
  hfom->SetName("fom");
  hfom->Fit("pol3","q0");
  fom  =hfom->GetFunction("pol3")->GetMaximum();
  alpha=hfom->GetFunction("pol3")->GetMaximumX();

  if (draw) {
    gROOT->SetStyle("Plain");
    gStyle->SetPalette(ipal);
    gStyle->SetOptTitle(0);
    gStyle->SetOptStat(0);

    TString xtitle=Form("#alpha = (#phi_{sf}-%g)/%g", vpar[0],dpar[0]);
    for (int j=0;j<nch;++j) {
      xtitle+=Form(" = (#phi_{%c}^{%c}-%g)/%g",
		   vch[j][1],(vch[j][0]=='f'?'+':'-'),
		   vpar[1+2*j],dpar[1+2*j]);
      xtitle+=Form(" = (#omega_{%c}^{%c}-%g)/%g",
		   vch[j][1],(vch[j][0]=='f'?'+':'-'),
		   vpar[2+2*j],dpar[2+2*j]);
    }
    if (ksp==0) xtitle="#phi_{sf} [deg]           ";
    if (ksp>0)  xtitle=Form("#%s_{%c}^{%s} [deg]           ",
			    ksp%2?"phi":"eta", vch[(ksp-1)/2][1],
			    vch[(ksp-1)/2][0]=='f'?'+':'-' );
    hp2n->GetXaxis()->SetTitle(xtitle);
    hp2n->GetYaxis()->SetTitle("t_{LH2 targ} / T_{pulse}           ");
    hp2n->Draw("col");

    TH1*  hsc = (TH1*)hfom->Clone("hsc");
    float hmin = hsc->GetMinimum();
    float hmax = hsc->GetMaximum();
    float margin = (hmax-hmin)/10;
    for (int i=0;i<=npt+1;i++) hsc->SetBinContent
      (i, (hsc->GetBinContent(i)-(hmin-margin))/(12*margin) );
    hsc->SetLineColor(icol);
    hsc->SetLineWidth(2);

    hsc->Draw("same");
    hsc->Fit("pol3",fit?"q":"q0","same");
    hsc->GetFunction("pol3")->SetLineWidth(2);
    hsc->GetFunction("pol3")->SetLineColor(kGreen);
    alpha=hsc->GetFunction("pol3")->GetMaximumX();

    TGaxis *axis = new TGaxis
      (max,0, max,1, hmin-margin, hmax+margin, 510, "+L");
    axis->SetLineColor(icol);
    axis->SetLabelColor(icol);
    axis->SetTextColor(icol);
    axis->SetTitle("integrated FOM           ");
    axis->Draw();
    gPad->Update();
  }
  cout<<"  [ ";
  for (int j=0;j<npar;++j)
    cout<<Form("%6.2f ",vpar[j]);
  cout<<Form("] -> %8.4f (%+6.3f)\n",fom,alpha);

  return hfom;
}


TFile* f;
TNtuple* ntp;

float  vpar[]={ 200,  73, 125, 98, 168 };
float  dpar[]={  25,   0,   0,  0,   0 };
int    npt, nev, nch = 2;
char*  vch[]={"f1","g2"};
int    ntof=500;
bool   draw=true;
int    fit=true;

TH1* hfom;

void p2n_line(int npt_=25, int nev_=1000000)
{
  npt=npt_; nev=nev_;
  f=new TFile("mcstas.root");
  ntp=(TNtuple*)f->Get("mcstas");
  hfom = p2n_line_fill(ntp,npt,nev,vpar,dpar,nch,vch,ntof,draw,fit);
}

void set(float alpha, float* dp=dpar)
{
  for (int j=0;j<1+2*nch;++j)
    vpar[j] += alpha * dp[j];
}

void sim(float* dp=dpar)
{
  hfom = p2n_line_fill(ntp,npt,nev,vpar,dp,nch,vch,ntof,draw,fit);
}

void sim(float x0, float x1=0,float x2=0, float x3=0,float x4=0,
                   float x5=0,float x6=0, float x7=0,float x8=0)
{
  float dset[]={x0, x1, x2, x3, x4, x5, x6, x7, x8};
  for (int j=0;j<1+2*nch;++j) dpar[j]=dset[j];
  sim();
}

// semi-orthogonal directions
float db[5][5] = {
  {5,0,0,0}, {0,5,0,5,0}, {0,0,5,0,5}, {0,5,0,-5,0}, {0,0,5,0,-5}
};

// optimize 2 choppers (5 parameters)
void opt5(int iter=1)
{
  for (int k=0;k<iter;++k) {
    cout<<"Iteration #"<<k<<endl;
    for (int j=0;j<5;++j) {
      sim(db[j]);
      if (alpha<-1) alpha=-1;
      if (alpha> 1) alpha= 1;
      set(alpha,db[j]);
    }
  }
}


double cf[]={  73,  98, 117, 136 };
double cw[]={ 125, 168, 201, 234 };
char vch1[3]="", vch2[3]="";


void opt_pair(int ch1=1, int ch2=2, int iter=8)
{
  char* vcc[4]={"ff","gg","fg","gf"};

  vch[0]=vch1;  vch[1]=vch2;
  vch1[1]='0'+ch1; vch2[1]='0'+ch2;

  // optimize each combo of choppers

  for (int i=0;i<6;++i) {

    // reset parameters
    vpar[0]=200;
    vpar[1]=cf[ch1-1]; vpar[2]=cw[ch1-1];
    vpar[3]=cf[ch2-1]; vpar[4]=cw[ch2-1];

    // set directions
    vch1[0]=vcc[i][0]; vch2[0]=vcc[i][1];
    cout<<endl<<"Choppers "<<vch1<<", "<<vch2<<endl<<endl;
    opt5(iter);

  }
}


void opt_combo(char* vcc="fg", int iter=8)
{
  // chopper combinations

  int ch1[6]={1,2,3,1,2,1};
  int ch2[6]={2,3,4,3,4,4};

  vch[0]=vch1;  vch[1]=vch2;
  vch1[0]=vcc[0]; vch2[0]=vcc[1];

  // optimize each combo of choppers

  for (int i=0;i<6;++i) {

    // reset parameters
    vpar[0]=200;
    vpar[1]=cf[ch1[i]-1]; vpar[2]=cw[ch1[i]-1];
    vpar[3]=cf[ch2[i]-1]; vpar[4]=cw[ch2[i]-1];

    vch1[1]='0'+ch1[i];   vch2[1]='0'+ch2[i];
    cout<<endl<<"Choppers "<<vch1<<", "<<vch2<<endl<<endl;
    opt5(iter);

  }
}

// optimize single chopper (3 parameters)

float ds[3][3] = {
  {5,0,0}, {0,5,0}, {0,0,5}
};

void opt_single(char* vcc="f", int iter=8)
{
  vch[0]=vch1; vch1[0]=vcc[0];

  // loop over chopper number
  nch=1;
  for (int i=0;i<4;++i) {

    // reset parameters
    vpar[0]=200; vpar[1]=cf[i]; vpar[2]=cw[i];
    vch1[1]='0'+i+1;
    cout<<endl<<"Chopper "<<vch1<<endl<<endl;

    for (int k=0;k<iter;++k) {
      cout<<"Iteration #"<<k<<endl;
      for (int j=0;j<3;++j) {
	sim(ds[j]);
	if (alpha<-1) alpha=-1;
	if (alpha> 1) alpha= 1;
	set(alpha,ds[j]);
      }
    }
  }
}
